@@ -24,8 +24,6 @@ from .. import mlog
 from ..environment import Environment
 from ..mesonlib import MachineChoice, version_compare
 from ..compilers.compilers import lang_suffixes, header_suffixes, obj_suffixes, lib_suffixes, is_header
-from subprocess import Popen, PIPE
-from threading import Thread
 from enum import Enum
 from functools import lru_cache
 from pathlib import Path
@@ -165,6 +163,14 @@ class OutputTargetMap:
     def target(self, name: str) -> T.Optional[T.Union['ConverterTarget', 'ConverterCustomTarget']]:
         return self._return_first_valid_key([self._target_key(name)])
 
+    def executable(self, name: str) -> T.Optional['ConverterTarget']:
+        tgt = self.target(name)
+        if tgt is None or not isinstance(tgt, ConverterTarget):
+            return None
+        if tgt.meson_func() != 'executable':
+            return None
+        return tgt
+
     def artifact(self, name: str) -> T.Optional[T.Union['ConverterTarget', 'ConverterCustomTarget']]:
         keys = []
         candidates = [name, OutputTargetMap.rm_so_version.sub('', name)]
@@ -661,7 +667,7 @@ class ConverterCustomTarget:
             for j in i:
                 if not j:
                     continue
-                target = output_target_map.target(j)
+                target = output_target_map.executable(j)
                 cmd += [target] if target else [j]
 
             commands += [cmd]
@@ -727,7 +733,7 @@ class ConverterCustomTarget:
             return None
 
     def log(self) -> None:
-        mlog.log('Custom Target', mlog.bold(self.name))
+        mlog.log('Custom Target', mlog.bold(self.name), '({})'.format(self.cmake_name))
         mlog.log('  -- command:      ', mlog.bold(str(self.command)))
         mlog.log('  -- outputs:      ', mlog.bold(str(self.outputs)))
         mlog.log('  -- conflict_map: ', mlog.bold(str(self.conflict_map)))
@@ -766,7 +772,7 @@ class CMakeInterpreter:
         self.languages = []
         self.targets = []
         self.custom_targets = []  # type: T.List[ConverterCustomTarget]
-        self.trace = CMakeTraceParser()
+        self.trace = CMakeTraceParser('', '')  # Will be replaced in analyse
         self.output_target_map = OutputTargetMap(self.build_dir)
 
         # Generated meson data
@@ -779,6 +785,7 @@ class CMakeInterpreter:
         cmake_exe = CMakeExecutor(self.env, '>=3.7', for_machine)
         if not cmake_exe.found():
             raise CMakeException('Unable to find CMake')
+        self.trace = CMakeTraceParser(cmake_exe.version(), self.build_dir, permissive=True)
 
         preload_file = Path(__file__).resolve().parent / 'data' / 'preload.cmake'
 
@@ -789,8 +796,8 @@ class CMakeInterpreter:
             preload_var = 'CMAKE_TOOLCHAIN_FILE'
 
         generator = backend_generator_map[self.backend_name]
-        cmake_args = cmake_exe.get_command()
-        trace_args = ['--trace', '--trace-expand', '--no-warn-unused-cli']
+        cmake_args = []
+        trace_args = self.trace.trace_args()
         cmcmp_args = ['-DCMAKE_POLICY_WARNING_{}=OFF'.format(x) for x in disable_policy_warnings]
         pload_args = ['-D{}={}'.format(preload_var, str(preload_file))]
 
@@ -830,46 +837,15 @@ class CMakeInterpreter:
             os.makedirs(self.build_dir, exist_ok=True)
             os_env = os.environ.copy()
             os_env['LC_ALL'] = 'C'
-            final_command = cmake_args + trace_args + cmcmp_args + pload_args + [self.src_dir]
-            proc = Popen(final_command, stdout=PIPE, stderr=PIPE, cwd=self.build_dir, env=os_env)
+            final_args = cmake_args + trace_args + cmcmp_args + pload_args + [self.src_dir]
 
-            def print_stdout():
-                while True:
-                    line = proc.stdout.readline()
-                    if not line:
-                        break
-                    mlog.log(line.decode('utf-8').strip('\n'))
-                proc.stdout.close()
-
-            t = Thread(target=print_stdout)
-            t.start()
-
-            # Read stderr line by line and log non trace lines
-            self.raw_trace = ''
-            tline_start_reg = re.compile(r'^\s*(.*\.(cmake|txt))\(([0-9]+)\):\s*(\w+)\(.*$')
-            inside_multiline_trace = False
-            while True:
-                line = proc.stderr.readline()
-                if not line:
-                    break
-                line = line.decode('utf-8')
-                if tline_start_reg.match(line):
-                    self.raw_trace += line
-                    inside_multiline_trace = not line.endswith(' )\n')
-                elif inside_multiline_trace:
-                    self.raw_trace += line
-                else:
-                    mlog.warning(line.strip('\n'))
-
-            proc.stderr.close()
-            proc.wait()
-
-            t.join()
+            cmake_exe.set_exec_mode(print_cmout=True, always_capture_stderr=self.trace.requires_stderr())
+            rc, _, self.raw_trace = cmake_exe.call(final_args, self.build_dir, env=os_env, disable_cache=True)
 
         mlog.log()
-        h = mlog.green('SUCCEEDED') if proc.returncode == 0 else mlog.red('FAILED')
+        h = mlog.green('SUCCEEDED') if rc == 0 else mlog.red('FAILED')
         mlog.log('CMake configuration:', h)
-        if proc.returncode != 0:
+        if rc != 0:
             raise CMakeException('Failed to configure the CMake subproject')
 
     def initialise(self, extra_cmake_options: T.List[str]) -> None:
@@ -924,7 +900,6 @@ class CMakeInterpreter:
         self.languages = []
         self.targets = []
         self.custom_targets = []
-        self.trace = CMakeTraceParser(permissive=True)
 
         # Parse the trace
         self.trace.parse(self.raw_trace)
@@ -1011,7 +986,7 @@ class CMakeInterpreter:
             if isinstance(value, str):
                 return string(value)
             elif isinstance(value, bool):
-                return BooleanNode(token(), value)
+                return BooleanNode(token(val=value))
             elif isinstance(value, int):
                 return number(value)
             elif isinstance(value, list):
@@ -1035,7 +1010,7 @@ class CMakeInterpreter:
             if not isinstance(args, list):
                 args = [args]
             args_n.arguments = [nodeify(x) for x in args if x is not None]
-            args_n.kwargs = {k: nodeify(v) for k, v in kwargs.items() if v is not None}
+            args_n.kwargs = {id_node(k): nodeify(v) for k, v in kwargs.items() if v is not None}
             func_n = FunctionNode(self.subdir, 0, 0, 0, 0, name, args_n)
             return func_n
 
@@ -1046,7 +1021,7 @@ class CMakeInterpreter:
             if not isinstance(args, list):
                 args = [args]
             args_n.arguments = [nodeify(x) for x in args if x is not None]
-            args_n.kwargs = {k: nodeify(v) for k, v in kwargs.items() if v is not None}
+            args_n.kwargs = {id_node(k): nodeify(v) for k, v in kwargs.items() if v is not None}
             return MethodNode(self.subdir, 0, 0, obj, name, args_n)
 
         def assign(var_name: str, value: BaseNode) -> AssignmentNode:
@@ -1157,7 +1132,7 @@ class CMakeInterpreter:
 
             # Generate target kwargs
             tgt_kwargs = {
-                'build_by_default': False,
+                'build_by_default': tgt.install,
                 'link_args': tgt.link_flags + tgt.link_libraries,
                 'link_with': link_with,
                 'include_directories': id_node(inc_var),
